package wrapper;

import java.util.HashSet;
import java.util.Set;

import org.neo4j.graphdb.Direction;
import org.neo4j.graphdb.DynamicRelationshipType;
import org.neo4j.graphdb.Node;
import org.neo4j.graphdb.Relationship;
import org.neo4j.graphdb.Transaction;
import org.neo4j.index.IndexHits;
import org.neo4j.index.lucene.LuceneIndexService;
import org.neo4j.kernel.EmbeddedGraphDatabase;

public class GraphWrapper {

	private LuceneIndexService index;
	private Transaction tx;
	private EmbeddedGraphDatabase graphDb;
	private int contatore=0;

	/**
	 * Apre il database Neo4j indicato.
	 * Il nome è il percorso, anche relativo, della cartella del
	 * database, con o senza la / finale
	 * */
	public GraphWrapper(String nome){
		System.out.println("apro il database");
		if (!nome.endsWith("/"))
			this.graphDb = new EmbeddedGraphDatabase( nome );
		else
			this.graphDb = new EmbeddedGraphDatabase( nome+"/" );

		System.out.println("avvio l'indice");
		this.index = new LuceneIndexService( graphDb );
		System.out.println("apro la transazione");
		this.tx = graphDb.beginTx();
	}

	/**
	 * Restituisce i nodi precedentementi indicizzati con indicizza()
	 * con etichetta e valore indicati
	 * */
	public IndexHits<Node> getNodes(String etichetta, Object valore) {
		return index.getNodes(etichetta, valore);
	}

	/**
	 * Chiude la transazione interna, l'indice e infine il database
	 * Va SEMPRE chiamato prima della chiusura dell'applicazione, altrimenti 
	 * il grafo o l'indice potrebbero subire dei danni
	 * */
	public void termina() {
		System.out.println("finisco");
		tx.success();
		tx.finish();
		System.out.println("chiudo il database");
		graphDb.shutdown();
		System.out.println("Fatto, termino");
		//System.exit(0);
	}

	public Node getCategoria(String nome) {
		for(Node n:index.getNodes("name", nome))
			if (n.hasProperty("idcat")) return n;

		return null;
	}

	/**
	 * Restituisce le categorie di appartenenza di una pagina o una categoria
	 * Segue sia le relazioni BELONGSTO che SUBCATEGORYOF
	 * */
	public Set<Node> getCategorieAppartenenza(Node n) {
		contatore++;
		if(contatore%100==0){
			tx.success();
			tx.finish();
			tx=graphDb.beginTx();
		}
		Set<Node> risultato=new HashSet<Node>();
		for(Relationship r:n.getRelationships(DynamicRelationshipType.withName("BELONGSTO"),Direction.OUTGOING)){
			risultato.add(r.getEndNode());
		}
		for(Relationship r:n.getRelationships(DynamicRelationshipType.withName("SUBCATEGORYOF"),Direction.OUTGOING)){
			risultato.add(r.getEndNode());
		}
		return risultato;

	}


	/**
	 * Restituisce le categorie contenute in una categoria data,
	 * seguendo le relazioni SUBCATEGORYOF
	 * */
	public Set<Node> getCategorieContenute(Node n) {
		contatore++;
		if(contatore%100==0){
			tx.success();
			tx.finish();
			tx=graphDb.beginTx();
		}
		Set<Node> risultato=new HashSet<Node>();
		for(Relationship r:n.getRelationships(DynamicRelationshipType.withName("SUBCATEGORYOF"))){
			risultato.add(r.getStartNode());
		}
		return risultato;

	}

	public Iterable<Node> getAllNodes() {
		return graphDb.getAllNodes();
	}

	/**
	 * Inserisce nell'indice il nodo dato assegnandoli un'etichetta e un valore dati
	 * @param n : nodo da indicizzare
	 * @param nome : nome dell'etichetta con cui indicizzarlo
	 * @param valore : valore da assegnare all'etichetta
	 * */
	public void indicizza(Node n,String nome, Object valore) {
		contatore++;

		if(contatore%100==0){
			tx.success();
			tx.finish();
			tx=graphDb.beginTx();
		}
		index.index(n,nome,valore);

	}

	/**
	 * Chiude e riapre la transazione usata dal wrapper
	 * Andrebbe chiamato a intervalli regolari se si manipola 
	 * il grafo al di fuori del wrapper
	 * */
	public void commit() {
		tx.success();
		tx.finish();
		tx=graphDb.beginTx();
	}

	/**
	 * Restituisce la pagina con un certo nome (è case sensitive)
	 * Se non esiste o non è indicizzata restituisce null
	 * */
	public Node getPagina(String nome) {
		for(Node n:index.getNodes("name", nome))
			if (n.hasProperty("idpag")) return n;

		return null;
	}

	/**
	 * Cancella tutte le relazioni da e verso un nodo e poi lo elimina
	 * Esegue anche un commit della transazione usata internamente dall'oggetto
	 * */
	public void elimina(Node el) {

		while(el.hasRelationship())
			for(Relationship r:el.getRelationships(Direction.BOTH)){
				System.out.println("elimino relazione "+r.getEndNode().getProperty("name")+" - "+r.getStartNode().getProperty("name"));
				r.delete();
				contatore++;
				if(contatore%100==0){
					//ogni 100 righe svuoto e ricomincio la transazione
					tx.success();
					tx.finish();
					tx=graphDb.beginTx();
				}
				//tentativo:interrompi il ciclo, che Ã¨ fasullo:non si itera su una cosa che si modifica
				break;

			}
		el.delete();
		commit();

	}

	/**
	 * Crea un nodo vuoto nel grafo e lo restituisce
	 * */
	public Node createNode(){
		contatore++;
		if(contatore%100==0){
			tx.success();
			tx.finish();
			tx=graphDb.beginTx();
		}

		return graphDb.createNode();
	}


	/**
	 * Assegna una categoria o una pagina a un'altra, creando una relazione tra le due
	 * la relazione sarà di tipo BELONGSTO o SUBCATEGORYOF a seconda dei casi
	 * */
	public void categorizzaIn(Node contenuto, Node contenitore) {
		contatore++;
		if(contatore%100==0){
			tx.success();
			tx.finish();
			tx=graphDb.beginTx();
		}

		if(contenuto.hasProperty("idcat"))
			contenuto.createRelationshipTo(contenitore, DynamicRelationshipType.withName("SUBCATEGORYOF"));
		else
			contenuto.createRelationshipTo(contenitore, DynamicRelationshipType.withName("BELONGSTO"));

	}

	/**
	 * Restituisce il set delle pagine contenute in una data categoria
	 * Se viene passato come argomento una pagina o una categoria vuota si ottiene un set vuoto
	 * */
	public HashSet<Node> getPagineContenute(Node n) {
		HashSet<Node> risultato=new HashSet<Node>();
		for(Relationship r:n.getRelationships(DynamicRelationshipType.withName("BELONGSTO"))){
			risultato.add(r.getStartNode());
		}
		return risultato;
	}


}
